from django.core.exceptions import ValidationError


class FreezeValidationError(ValidationError):
    """Data and field validation-related error."""

    pass


class FreezeConfigurationError(ValidationError):
    """Configuration-related error."""

    pass



import threading
from collections import defaultdict
from contextlib import contextmanager
from typing import Any, Iterable, Optional, Union

from dirtyfields import DirtyFieldsMixin
from django.core.exceptions import FieldDoesNotExist
from django.db import models
from django.db.models.signals import class_prepared
from django.dispatch import receiver
from django_fsm import FSMField

from django_fsm_freeze.exceptions import (
    FreezeConfigurationError,
    FreezeValidationError,
)

_DISABLED_FSM_FREEZE = threading.local()


@contextmanager
def bypass_fsm_freeze(
    objs: Union[
        'FreezableFSMModelMixin', Iterable['FreezableFSMModelMixin']
    ] = (),
    bypass_globally: bool = False,
):
    """
    Bypass the frozen checks.

    objs: the object(s) that will not be checked for its frozeness
    bypass_globally: flag to apply the bypassing globally
    """

    if objs and not isinstance(objs, Iterable):
        objs = (objs,)
    errors = []
    for obj in objs:
        if not isinstance(obj, FreezableFSMModelMixin):
            errors.append(
                f'Unsupported argument(s): {obj!r}. '
                f'`bypass_fsm_freeze()` accepts instance(s) from '
                f'FreezableFSMModelMixin.'
            )
    if errors:
        raise FreezeConfigurationError(errors)

    try:
        if bypass_globally is True:
            _DISABLED_FSM_FREEZE.active = True
        for obj in objs:
            obj._bypass_fsm_freeze = True
        yield
    finally:
        if bypass_globally is True:
            _DISABLED_FSM_FREEZE.active = False
        for obj in objs:
            obj._bypass_fsm_freeze = False


def resolve_dotted_path(instance: Any, path: str) -> Any:
    """
    Walk recursively the path separated by dots.
    """
    for part in path.split('.'):
        instance = getattr(instance, part)
    return instance


class FreezableFSMModelMixin(DirtyFieldsMixin, models.Model):
    """
    Support for django-fsm data immutability.

    FROZEN_IN_STATES: fsm states that an instance is considered frozen in
    FROZEN_STATE_LOOKUP_FIELD: the field used to be looked up for the state
    FROZEN_DELEGATE_TO: the field used to be looked up for the state (via
                        foreignkey, dot-separated path).
                        Cannot be combined with `FROZEN_STATE_LOOKUP_FIELD`.
    NON_FROZEN_FIELDS: fields that are mutable
    """

    class Meta:
        abstract = True

    FROZEN_IN_STATES: tuple = ()
    FROZEN_STATE_LOOKUP_FIELD: Optional[str]
    FROZEN_DELEGATE_TO: Optional[str] = None
    NON_FROZEN_FIELDS: tuple = ()

    _bypass_fsm_freeze: bool = False

    @property
    def is_fsm_frozen(self) -> bool:
        """Determine whether self is frozen or not."""

        instance, fsm_field = self._resolve_delegation()
        return (
            fsm_field.value_from_object(instance) in instance.FROZEN_IN_STATES
        )

    def _resolve_delegation(self) -> tuple['FreezableFSMModelMixin', FSMField]:
        """
        Find the FreezableFSMModelMixin instance and its FSMState field
        """

        if self.FROZEN_DELEGATE_TO:
            instance = resolve_dotted_path(self, self.FROZEN_DELEGATE_TO)
            if not isinstance(instance, FreezableFSMModelMixin):
                raise FreezeConfigurationError(
                    {
                        'FROZEN_DELEGATE_TO': [
                            'Does not resolve to a'
                            ' FreezableFSMModelMixin model.'
                        ]
                    }
                )
            return instance, instance.__class__._get_fsm_field()
        return self, self.__class__._get_fsm_field()

    @property
    def _is_fsm_freeze_bypassed(self) -> bool:
        return bool(
            getattr(_DISABLED_FSM_FREEZE, 'active', False)
            or self._bypass_fsm_freeze
        )

    def freeze_check(self) -> None:
        """Check dirty fields and frozen status.

        Raise `FreezeValidationError` if it is dirty and frozen.
        """

        if self._is_fsm_freeze_bypassed or not self.is_fsm_frozen:
            return
        errors = defaultdict(list)
        instance, fsm_field = self._resolve_delegation()
        dirty_fields = self.get_dirty_fields(check_relationship=True)
        for field in set(dirty_fields) - set(
            instance.NON_FROZEN_FIELDS + (fsm_field.name,)
        ):
            errors[field].append('Cannot change frozen field.')
        if errors:
            raise FreezeValidationError(errors)

    @classmethod
    def _get_fsm_field(cls) -> FSMField:
        """Discover the FSMField.

        If multiples are found, we use FROZEN_STATE_LOOKUP_FIELD to select it.
        """

        fsm_fields = [
            field for field in cls._meta.fields if isinstance(field, FSMField)
        ]
        if not fsm_fields:
            raise FieldDoesNotExist
        if len(fsm_fields) == 1:
            # Autodetected
            return fsm_fields[0]
        if not hasattr(cls, 'FROZEN_STATE_LOOKUP_FIELD'):
            raise TypeError(
                'Ambiguity to find the frozen state lookup field.'
                ' Please define FROZEN_STATE_LOOKUP_FIELD attribute'
                f' on the class {cls!r}'
            )
        for field in fsm_fields:
            if field.name == cls.FROZEN_STATE_LOOKUP_FIELD:
                return field
        raise FieldDoesNotExist

    @classmethod
    def config_check(cls) -> None:
        errors = defaultdict(list)
        if cls.FROZEN_DELEGATE_TO:
            if cls.FROZEN_IN_STATES:
                errors['FROZEN_IN_STATES'].append(
                    'Field FROZEN_DELEGATE_TO is already defined.'
                )
            if getattr(cls, 'FROZEN_STATE_LOOKUP_FIELD', None):
                errors['FROZEN_STATE_LOOKUP_FIELD'].append(
                    'Field FROZEN_DELEGATE_TO is already defined.'
                )
        else:
            try:
                cls._get_fsm_field()
            except FieldDoesNotExist:
                errors['FROZEN_STATE_LOOKUP_FIELD'].append(
                    'FSMField not found.'
                )
            except TypeError as err:
                errors['FROZEN_STATE_LOOKUP_FIELD'].append(str(err))

        for field in cls.NON_FROZEN_FIELDS:
            try:
                cls._meta.get_field(field)
            except FieldDoesNotExist:
                errors[field].append(f'{field!r} field does not exist.')
        if errors:
            raise FreezeConfigurationError(errors)

    def save(self, *args, **kwargs) -> None:
        """Data freeze checking before saving the object."""

        if not kwargs.get('force_insert', None):
            # e.g. not object creation
            self.freeze_check()

        return super().save(*args, **kwargs)

    def delete(self, *args, **kwargs):
        if not self._is_fsm_freeze_bypassed and self.is_fsm_frozen:
            raise FreezeValidationError(
                f'{self!r} is frozen, cannot be deleted.'
            )
        return super().delete(*args, **kwargs)


@receiver(class_prepared)
def on_class_prepared(sender, **kwargs):
    if issubclass(sender, FreezableFSMModelMixin):
        sender.config_check()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['django_fsm_freeze']

package_data = \
{'': ['*']}

install_requires = \
['Django', 'django-dirtyfields>=1.7.0,<2.0.0', 'django-fsm']

setup_kwargs = {
    'name': 'django-fsm-freeze',
    'version': '0.1.9',
    'description': 'django-fsm data immutability support',
    'long_description': "# django fsm data immutability support\n![CI](https://github.com/ming-tung/django-fsm-freeze/actions/workflows/continues-integration.yml/badge.svg?branch=main)\n[![PyPI version](https://badge.fury.io/py/django-fsm-freeze.svg)](https://badge.fury.io/py/django-fsm-freeze)\n[![Downloads](https://static.pepy.tech/personalized-badge/django-fsm-freeze?period=total&units=international_system&left_color=grey&right_color=yellowgreen&left_text=Downloads)](https://pepy.tech/project/django-fsm-freeze)\n\ndjango-fsm-freeze provides a django model mixin for data immutability based on\n[django-fsm](https://github.com/viewflow/django-fsm).\n\n\n## Installation\n\n```commandline\npip install django-fsm-freeze\n```\n\n## Configuration\n\n### Basic configuration\n- Add `FreezableFSMModelMixin` to your [django-fsm](https://github.com/viewflow/django-fsm) model\n- Specify the `FROZEN_IN_STATES` in which the object should be frozen, meaning the\n  value of its fields/attributes cannot be changed.\n- (optional) Customize the `NON_FROZEN_FIELDS` for partial mutability\n\nWhen an object is in a frozen state, by default all of its fields are immutable,\nexcept for the `state` FSMField which needs to be mutable for\n[django-fsm](https://github.com/viewflow/django-fsm) to work.\n\n```python\nfrom django_fsm import FSMField\n\nfrom django_fsm_freeze.models import FreezableFSMModelMixin\n\nclass MyDjangoFSMModel(FreezableFSMModelMixin):\n\n    # In this example, when object is in the 'active' state, it is immutable.\n    FROZEN_IN_STATES = ('active', )\n    \n    # django-fsm specifics: state, transitions, etc.\n    state = FSMField(default='new')\n    # ...\n```\n\n### Customization\n\n#### Tell django-fsm-freeze which field to look up for frozeness\nBy default, FreezableFSMModelMixin will look for the FSMField on your model\nand its value to determine whether the instance is frozen or not.\nHowever, in case your model has multiple `FSMField`s,\nyou would need to tell the Mixin which field should be used to look up,\nto determine the frozeness via the `FROZEN_STATE_LOOKUP_FIELD` attribute.\n\n```python\nfrom django.db import models\nfrom django_fsm import FSMField\n\nfrom django_fsm_freeze.models import FreezableFSMModelMixin\n\nclass MyDjangoFSMModel(FreezableFSMModelMixin):\n\n    # In this example, when object is in the 'active' state, it is immutable.\n    FROZEN_IN_STATES = ('active', )\n\n    # Assign this with the name of the `FSMField` if your models has multiple FSMFields.\n    # See example in `mytest/models.py:FakeModel2`\n    FROZEN_STATE_LOOKUP_FIELD = 'state'\n    \n    # django-fsm specifics: state, transitions, etc.\n    state = FSMField(default='new')\n    another_state = FSMField(default='draft')\n    # ...\n```\n\nIn another case, when the desired lookup state is on another model related\nvia foreign key, instead of setting `FROZEN_STATE_LOOKUP_FIELD`,\nit is possible to specify the (dot-separated) path to that field in\n`FROZEN_DELEGATE_TO`.\nThis setting instructs the freezable model instance to evaluate the freezable\nstate from that remote field.\n\n```python\nclass Parent(FreezableFSMModelMixin):\n    state = FSMField(default='new')\n\n\nclass Child(FreezableFSMModelMixin):\n\n    # Assign this with the path (dotted separated) to the instance you expect\n    # the decision for freezability to be decided on.\n    FROZEN_DELEGATE_TO = 'parent'\n    parent = models.ForeignKey(Parent, on_delete=models.PROTECT)\n```\n\n#### Define for partial mutability \nIn case we want to mutate certain fields when the object is frozen, we can\nset the `NON_FROZEN_FIELDS` to allow it.\n\n```python\nclass MyDjangoFSMModel(FreezableFSMModelMixin):\n\n    # In this example, when object is in the 'active' state, it is immutable.\n    FROZEN_IN_STATES = ('active', )\n    NON_FROZEN_FIELDS = ('a_mutable_field', )\n\n    # This field is mutable even when the object is in the frozen state.\n    a_mutable_field = models.BooleanField()\n```\nSee configuration example in https://github.com/ming-tung/django-fsm-freeze/blob/main/mytest/models.py\n\n## Usage\n\nThe frozen check takes place when\n - class is prepared (configuration checking)\n - `object.save()`\n - `object.delete()`\n\nIn case of trying to save/delete a frozen object, a `FreezeValidationError` will be raised.\nIn case of misconfiguration, a `FreezeConfigurationError` will be raised.\n\n\n### Bypassing\nIf you want to bypass the frozen check for some reason, you can use the contextmanager\n`bypass_fsm_freeze()`, with the freezable object(s) that you want to bypass\nthe checks on, or apply the bypass globally via `bypass_globally` argument.\n\nYou can find some usage example in test `mytest/test_models.py:TestBypassFreezeCheck`.\n\n## Developing\nFor contributors or developers of the project, please see [DEVELOPING.md](docs/DEVELOPING.md)\n\n## Contributing \n(TODO)\nFor anyone who is interested in contributing to this project, please see [CONTRIBUTING.md](docs/CONTRIBUTING.md).\nThank you :)\n\nFor further discussions or suggestions, you could also reach out to me on twitter or email.\n",
    'author': 'ming-tung',
    'author_email': 'mingtung.hong@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ming-tung/django-fsm-freeze',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
